package at.ac.uibk.gitsearch.repository.search;

import at.ac.uibk.gitsearch.es.model.DocumentInfo;
import at.ac.uibk.gitsearch.service.dto.GitFilesAggregationDTO;
import at.ac.uibk.gitsearch.service.dto.GitFilesDTO;
import at.ac.uibk.gitsearch.service.dto.GitFilesPageDetailsDTO;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.elasticsearch.action.search.SearchRequest;
import org.elasticsearch.action.search.SearchResponse;
import org.elasticsearch.client.RequestOptions;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.index.query.BoolQueryBuilder;
import org.elasticsearch.index.query.QueryBuilders;
import org.elasticsearch.search.SearchHit;
import org.elasticsearch.search.aggregations.AggregationBuilders;
import org.elasticsearch.search.aggregations.bucket.terms.TermsAggregationBuilder;
import org.elasticsearch.search.builder.SearchSourceBuilder;
import org.elasticsearch.search.fetch.subphase.highlight.HighlightBuilder;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Repository;

@Repository
public class GitFilesRepositoryImpl implements GitFilesRepository {

    @Value("${application.search.highlight-pre}")
    private String preTags;

    @Value("${application.search.highlight-post}")
    private String postTags;

    private final RestHighLevelClient elasticsearchClient;

    private static final int FRAGMENT_SIZE = 200;

    public GitFilesRepositoryImpl(RestHighLevelClient elasticsearchClient) {
        this.elasticsearchClient = elasticsearchClient;
    }

    @SuppressWarnings("deprecation")
    @Override
    public GitFilesPageDetailsDTO pageDetails(SearchInputDTO searchInputDTO) throws IOException {
        Collection<String> projectIDs = searchProjectsIDsMetadata(searchInputDTO);
        SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_FULLTEXT);

        BoolQueryBuilder queryBuilder = QueryBuilders
            .boolQuery()
            .must(
                QueryBuilders.simpleQueryStringQuery(searchInputDTO.getFulltextQuery()).field(SearchRepositoryConstants.FULLTEXT_CONTENT)
            );

        addFulltextSelection(queryBuilder, searchInputDTO.getSelectedRepository(), SearchRepositoryConstants.FULLTEXT_REPOSITORY);
        addFulltextSelection(queryBuilder, searchInputDTO.getSelectedRepository(), SearchRepositoryConstants.FULLTEXT_SUB_GROUP);
        addFulltextSelection(queryBuilder, searchInputDTO.getSelectedFileFormat(), SearchRepositoryConstants.FULLTEXT_FILE_EXTENSION);
        if (projectIDs != null) {
            queryBuilder.filter(QueryBuilders.termsQuery(SearchRepositoryConstants.FULLTEXT_PROJECT_ID, projectIDs));
        }

        char[] boundaryChars = { '\n' };

        HighlightBuilder highlightBuilder = new HighlightBuilder()
            .highlighterType("plain")
            .preTags(preTags)
            .postTags(postTags)
            .field(SearchRepositoryConstants.FULLTEXT_CONTENT)
            .boundaryChars(boundaryChars)
            .boundaryScannerType(HighlightBuilder.BoundaryScannerType.SENTENCE)
            .fragmentSize(FRAGMENT_SIZE);

        //String[] includes = {};
        //String[] excludes = {HighlightRepositoryConstants.CONTENT_FIELD};

        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder();
        sourceBuilder
            .query(queryBuilder)
            .highlighter(highlightBuilder)
            .size(searchInputDTO.getPageSize())
            .from((searchInputDTO.getPage() - 1) * searchInputDTO.getPageSize());

        searchRequest.source(sourceBuilder);

        SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

        ObjectMapper objectMapper = new ObjectMapper();
        objectMapper.registerModule(new JavaTimeModule());
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        List<GitFilesDTO> gitFilesDTOS = new ArrayList<>();
        for (SearchHit searchHit : searchResponse.getHits().getHits()) {
            DocumentInfo documentInfo = objectMapper.readValue(searchHit.getSourceAsString(), DocumentInfo.class);
            gitFilesDTOS.add(new GitFilesDTO(documentInfo, searchHit, preTags, postTags));
        }
        long hitCount = searchResponse.getHits().getTotalHits() == null ? 0 : searchResponse.getHits().getTotalHits().value;
        return new GitFilesPageDetailsDTO(gitFilesDTOS, hitCount);
    }

    private static void addFulltextSelection(BoolQueryBuilder queryBuilder, List<String> selectedFileFormat, String fileExtensionField) {
        if (selectedFileFormat.isEmpty()) {
            return;
        }
        queryBuilder.must(QueryBuilders.termsQuery(fileExtensionField, selectedFileFormat));
    }

    @Override
    public GitFilesAggregationDTO aggregation(String query) throws IOException {
        SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_FULLTEXT);

        BoolQueryBuilder matchQueryBuilder = QueryBuilders
            .boolQuery()
            .must(QueryBuilders.simpleQueryStringQuery(query).field(SearchRepositoryConstants.FULLTEXT_CONTENT));

        TermsAggregationBuilder groupSubGroup = AggregationBuilders
            .terms(SearchRepositoryConstants.FULLTEXT_SUB_GROUP_AGG)
            .field(SearchRepositoryConstants.FULLTEXT_SUB_GROUP);

        TermsAggregationBuilder groupFileExtension = AggregationBuilders
            .terms(SearchRepositoryConstants.FULLTEXT_FILE_EXTENSION_AGG)
            .field(SearchRepositoryConstants.FULLTEXT_FILE_EXTENSION);

        TermsAggregationBuilder groupRepository = AggregationBuilders
            .terms(SearchRepositoryConstants.FULLTEXT_REPOSITORY_AGG)
            .field(SearchRepositoryConstants.FULLTEXT_REPOSITORY);

        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder()
            .query(matchQueryBuilder)
            .size(0)
            .aggregation(groupSubGroup)
            .aggregation(groupFileExtension)
            .aggregation(groupRepository);

        searchRequest.source(sourceBuilder);

        SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);

        return new GitFilesAggregationDTO(searchResponse.getAggregations());
    }

    private Collection<String> searchProjectsIDsMetadata(SearchInputDTO searchInputDTO) throws IOException {
        if (!searchInputDTO.hasMetadataInput()) {
            return Collections.emptyList();
        }
        SearchRequest searchRequest = new SearchRequest(SearchRepositoryConstants.INDEX_METADATA);
        BoolQueryBuilder queryBuilder = QueryBuilders.boolQuery();
        addMetadataQueryShould(
            queryBuilder,
            searchInputDTO.getMetadata().getProgrammingLanguage(),
            SearchRepositoryConstants.METADATA_PROGRAMMING_LANGUAGES
        );
        addMetadataQueryShould(queryBuilder, searchInputDTO.getMetadata().getKeyword(), SearchRepositoryConstants.METADATA_KEYWORDS);
        addMetadataQueryMust(queryBuilder, searchInputDTO.getMetadata().getAuthor(), SearchRepositoryConstants.METADATA_CREATOR);
        addMetadataQueryMust(queryBuilder, searchInputDTO.getMetadata().getLicense(), SearchRepositoryConstants.METADATA_LICENSE);
        SearchSourceBuilder sourceBuilder = new SearchSourceBuilder()
            .query(queryBuilder)
            .fetchSource(SearchRepositoryConstants.METADATA_PROJECT_ID, null);
        // TODO: Allow more than 10 entries.
        searchRequest.source(sourceBuilder);
        SearchResponse searchResponse = elasticsearchClient.search(searchRequest, RequestOptions.DEFAULT);
        Collection<String> buckets = new ArrayList<>();
        for (SearchHit searchHit : searchResponse.getHits().getHits()) {
            buckets.add(searchHit.getId());
        }
        return buckets;
    }

    private static void addMetadataQueryShould(BoolQueryBuilder queryBuilder, String value, String field) {
        if (value == null) {
            return;
        }
        queryBuilder.should(QueryBuilders.termsQuery(field, value));
    }

    private static void addMetadataQueryMust(BoolQueryBuilder queryBuilder, String value, String field) {
        if (value == null) {
            return;
        }
        queryBuilder.must(QueryBuilders.simpleQueryStringQuery(value).analyzeWildcard(true).field(field));
    }
}
