package at.ac.uibk.gitsearch.service;

import static at.ac.uibk.gitsearch.service.MonitoringService.HOURLY_TAG;
import static at.ac.uibk.gitsearch.service.MonitoringService.NUMBER_DOWNLOAD_NAME;
import static at.ac.uibk.gitsearch.service.MonitoringService.NUMBER_EXERCISE_NAME;
import static at.ac.uibk.gitsearch.service.MonitoringService.NUMBER_LIKE_NAME;
import static at.ac.uibk.gitsearch.service.MonitoringService.NUMBER_USER_NAME;
import static at.ac.uibk.gitsearch.service.MonitoringService.NUMBER_VIEW_NAME;
import static at.ac.uibk.gitsearch.service.MonitoringService.USER_TAG;

import at.ac.uibk.gitsearch.IntegrationTest;
import at.ac.uibk.gitsearch.domain.Likes;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.repository.jpa.StatisticsRepository;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.service.dto.AdminUserDTO;
import at.ac.uibk.gitsearch.service.dto.StatisticsDTO;
import io.micrometer.core.instrument.Gauge;
import io.micrometer.core.instrument.Measurement;
import io.micrometer.core.instrument.MeterRegistry;
import io.micrometer.core.instrument.Tag;
import io.micrometer.core.instrument.Tags;
import io.micrometer.core.instrument.search.RequiredSearch;
import io.micrometer.prometheus.PrometheusMeterRegistry;
import java.io.IOException;
import java.time.Instant;
import java.time.LocalDate;
import java.util.Random;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import org.elasticsearch.node.NodeValidationException;
import org.junit.Assert;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.security.test.context.support.WithAnonymousUser;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.transaction.annotation.Transactional;

@IntegrationTest
@WithMockUser(value = "test", authorities = "sharing")
public class MonitoringServiceIT {

    @SuppressWarnings("unused")
    private static final Logger LOGGER = LoggerFactory.getLogger(MonitoringServiceIT.class);

    @Autowired
    private MonitoringService monitoringService;

    @Autowired
    private StatisticsService statisticsService;

    @Autowired
    private StatisticsRepository statisticsRepository;

    @Autowired
    PrometheusMeterRegistry hourlyMeterRegistry;

    @Autowired
    private LikesService likesService;

    @Autowired
    private UserService userService;

    @Autowired
    MeterRegistry meterRegistry;

    @Autowired
    private SearchService searchService;

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // cleanup all likes
    @AfterEach // before and after
    @Transactional
    public void cleanUpLikesEtAll() {
        likesService.findAll().stream().forEach(l -> likesService.delete(l.getId()));
        statisticsRepository.findAll().stream().forEach(s -> statisticsRepository.delete(s));
    }

    @BeforeEach // must be started as BeforeEach, in order to autowire the ElasticSearchTestConfiguration
    @Timeout(value = 3, unit = TimeUnit.MINUTES)
    public void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    @AfterAll
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    @Test
    @Timeout(value = 5, unit = TimeUnit.MINUTES)
    @WithAnonymousUser
    public void testMonitorHourlyMetrics() throws IOException {
        Random random = new Random(123L);

        AdminUserDTO user = new AdminUserDTO();
        user.setFirstName("first");
        user.setLastName("last");
        user.setEmail("abc@de.de");
        user.setLogin("someUser");
        user.setLastLogin(Instant.now());
        user.setActivated(true);
        User u = userService.createUser(user);

        AtomicInteger likes = new AtomicInteger(0);
        AtomicInteger downloads = new AtomicInteger(0);
        AtomicInteger views = new AtomicInteger(0);
        AtomicInteger exerciseCount = new AtomicInteger(0);

        searchService
            .getAllResultsAsStream()
            .forEach(ex -> {
                exerciseCount.addAndGet(1);
                StatisticsDTO stats = new StatisticsDTO();
                stats.setExerciseID(ex.getExerciseId());
                stats.setViews(random.nextInt(100));
                stats.setDownloads(random.nextInt(100));
                statisticsService.save(stats);
                downloads.addAndGet(stats.getDownloads());
                views.addAndGet(stats.getViews());

                if (random.nextBoolean()) {
                    Likes like = new Likes();
                    like.date(LocalDate.now()).exerciseID(ex.getExerciseId()).userID(u.getId().intValue());
                    likesService.save(like);
                    likes.incrementAndGet();
                }
            });

        monitoringService.monitorHourlyMetrics();

        // Countercheck results against original list
        Assert.assertEquals(
            "User counter may fail, if test data changes",
            5.0d,
            getLastMeterValue(NUMBER_USER_NAME, Tags.of(HOURLY_TAG, Tag.of(USER_TAG, "All"))),
            0.0001d
        );
        Assert.assertEquals(
            "Activae User counter may fail, if test data changes",
            0.0d,
            getLastMeterValue(NUMBER_USER_NAME, Tags.of(HOURLY_TAG, Tag.of(USER_TAG, "Active"))),
            0.0001d
        );
        Assert.assertEquals(
            "likes counter may fail, if test data changes",
            likes.doubleValue(),
            getLastMeterValue(NUMBER_LIKE_NAME),
            0.0001d
        );
        Assert.assertEquals(
            "View counter may fail, if test data changes",
            views.doubleValue(),
            getLastMeterValue(NUMBER_VIEW_NAME),
            1.0001d
        );
        Assert.assertEquals(
            "Download counter may fail, if test data changes",
            downloads.doubleValue(),
            getLastMeterValue(NUMBER_DOWNLOAD_NAME),
            1.0001d
        );
        Assert.assertEquals(
            "Exercise counter may fail, if test data changes",
            exerciseCount.doubleValue(),
            getLastMeterValue(NUMBER_EXERCISE_NAME),
            0.0001d
        );
        Assert.assertEquals(
            "Language counter may fail, if test data changes",
            11.0d,
            getLastMeterValue(NUMBER_EXERCISE_NAME, Tags.of(HOURLY_TAG, Tag.of(MonitoringService.LANGUAGE_TAG, "JAVA"))),
            0.0001d
        );
        Assert.assertEquals(
            "Language counter may fail, if test data changes",
            5.0d,
            getLastMeterValue(NUMBER_EXERCISE_NAME, Tags.of(HOURLY_TAG, Tag.of(MonitoringService.LANGUAGE_TAG, "Unclassified"))),
            0.0001d
        );
    }

    public double getLastMeterValue(String name) {
        final Gauge gauge = RequiredSearch.in(hourlyMeterRegistry).name(name).gauge();
        final Measurement next = gauge.measure().iterator().next();
        //        System.out.println(next);
        return next.getValue();
    }

    public double getLastMeterValue(String name, Tags tags) {
        final Gauge gauge = RequiredSearch.in(hourlyMeterRegistry).name(name).tags(tags).gauge();

        return gauge.value();
    }
}
