package at.ac.uibk.gitsearch.web.rest;

import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.domain.Review;
import at.ac.uibk.gitsearch.domain.ReviewRating;
import at.ac.uibk.gitsearch.domain.Statistics;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRatingRepository;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRepository;
import at.ac.uibk.gitsearch.repository.jpa.StatisticsRepository;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.ReviewService;
import at.ac.uibk.gitsearch.service.UserService;
import at.ac.uibk.gitsearch.service.dto.ReviewDTO;
import at.ac.uibk.gitsearch.service.dto.ReviewRequest;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.ObjectMapper;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.junit.jupiter.MockitoExtension;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;
import org.springframework.test.web.servlet.request.MockMvcRequestBuilders;

@SpringBootTest(classes = GitsearchApp.class)
@ExtendWith(MockitoExtension.class)
@AutoConfigureMockMvc
@WithMockUser
public class ReviewResourceIT {

    @Autowired
    private MockMvc mockMvc;

    @Autowired
    private ReviewRepository reviewRepository;

    @Autowired
    private ReviewRatingRepository reviewRatingRepository;

    @Autowired
    private ReviewService reviewService;

    @Autowired
    private UserService userService;

    @Autowired
    private StatisticsRepository statisticsRepository;

    @BeforeEach
    public void tearDown() {
        reviewRatingRepository.deleteAll();
        reviewRepository.deleteAll();
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    public void createReviewTest() throws Exception {
        ReviewDTO reviewDTO = new ReviewDTO();
        reviewDTO.setId(1L);
        reviewDTO.setResource("resource");
        User user = userService.getUserWithAuthorities().get();

        List<String> reviewers = new ArrayList<>();
        reviewers.add("admin@localhost");

        reviewDTO.setUsers(reviewers);

        reviewDTO.setResource("resource");

        reviewDTO.setBadgeRewarded(false);

        List<String> comm = new ArrayList<>();
        comm.add("comment");

        reviewDTO.setComments(comm);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/review/create")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        // Validate the Revoew is in the database
        List<Review> reviewList = reviewRepository.findAll();
        assertThat(reviewList).hasSize(1);
        Review testReview = reviewList.get(0);
        assertThat(testReview.getResource()).isEqualTo("resource");
        assertThat(testReview.getRequestedBy()).isEqualTo(user.getId());

        ReviewRating reviewRating = reviewRatingRepository.findAll().get(0);
        assertThat(reviewRating.getUser().getEmail()).isEqualTo("admin@localhost");
        assertThat(reviewRating.getReview().getResource()).isEqualTo("resource");
        assertThat(reviewRating.getStatus()).isEqualTo("Review In Progress");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    public void updateReviewTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        // Validate the Revoew is in the database
        List<Review> reviewList = reviewRepository.findAll();
        assertThat(reviewList).hasSize(1);

        Review testReview = reviewList.get(0);
        assertThat(testReview.getResource()).isEqualTo("resource");
        testReview.setResource("resource2");

        ReviewDTO reviewDTO = reviewService.reviewToDTO(testReview);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .put("/api/review")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        assertThat(reviewService.getReview(testReview.getId()).getResource()).isEqualTo("resource2");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void updateReviewRatingByUserTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));
        // Validate the Revoew is in the database
        List<ReviewRating> reviewList = reviewRatingRepository.findAllByUserId(userService.getUserWithAuthorities().get().getId());
        assertThat(reviewList).hasSize(1);

        ReviewRating testReview = reviewList.get(0);
        assertThat(testReview.getReview().getResource()).isEqualTo("resource");
        assertThat(testReview.getStatus()).isEqualTo("Review In Progress");

        ReviewDTO reviewDTO = reviewService.reviewToDTO(testReview.getReview());
        List<String> status = new ArrayList<>();
        status.add("DELETION REQUESTED");

        reviewDTO.setStatus(status);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .put("/api/review/updateByUser")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        List<ReviewDTO> reviews = reviewService.getAllReviews();

        assertEquals(0, reviews.size());
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    public void getReviewTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        List<Review> reviewList = reviewRepository.findAll();
        Review testReview = reviewList.get(0);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .get("/api/review/{id}", testReview.getId())
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        ReviewDTO reviewDTO = new ObjectMapper().readValue(str, ReviewDTO.class);

        assertThat(reviewDTO.getResource()).isEqualTo("resource");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    public void getAllReviewsTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));
        reviewService.createReview(new ReviewRequest("resource2", "codeability2.at", "2", reviewers));

        MvcResult result = mockMvc
            .perform(MockMvcRequestBuilders.get("/api/review/all").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON))
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        List<ReviewDTO> reviewDTOs = new ObjectMapper().readValue(str, new TypeReference<List<ReviewDTO>>() {});

        assertThat(reviewDTOs.size()).isEqualTo(2);
        assertThat(reviewDTOs.get(0).getResource()).isEqualTo("resource");
        assertThat(reviewDTOs.get(1).getResource()).isEqualTo("resource2");
        assertThat(reviewDTOs.get(0).getRequestedBy()).isEqualTo("admin@localhost");
        assertThat(reviewDTOs.get(1).getRequestedBy()).isEqualTo("admin@localhost");
        assertThat(reviewDTOs.get(0).getStatus().get(0)).isEqualTo("Review In Progress");
        assertThat(reviewDTOs.get(1).getStatus().get(0)).isEqualTo("Review In Progress");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    public void deleteReviewTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        List<Review> reviewList = reviewRepository.findAll();

        Review testReview = reviewList.get(0);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .delete("/api/review/{id}", testReview.getId())
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        assertThat(reviewRepository.findAll().size()).isEqualTo(0);
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void getReviewsByUser() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("user@localhost");
        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        MvcResult result = mockMvc
            .perform(MockMvcRequestBuilders.get("/api/review/allByUser").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON))
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        List<ReviewDTO> reviewDTOs = new ObjectMapper().readValue(str, new TypeReference<List<ReviewDTO>>() {});

        assertThat(reviewDTOs.get(0).getResource()).isEqualTo("resource");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void requestReviewTest() throws Exception {
        ReviewRequest reviewRequest = new ReviewRequest("resource", "codeability.at", "1", new ArrayList<>());

        User user = userService.getUserWithAuthorities().get();

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/requestReview")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewRequest))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        List<Review> reviewList = reviewRepository.findAll();

        Review testReview = reviewList.get(0);

        assertThat(testReview.getResource()).isEqualTo("resource");
        assertThat(testReview.getRequestedBy()).isEqualTo(user.getId());
        assertThat(testReview.getComments().size()).isZero();
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void notifyReviewersTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("admin@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        User user = userService.getUserWithAuthorities().get();

        List<Review> reviewList = reviewRepository.findAll();

        Review testReview = reviewList.get(0);

        ReviewDTO reviewDTO = reviewService.reviewToDTO(testReview);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/review/notifyReviewers")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        reviewList = reviewRepository.findAll();

        testReview = reviewList.get(0);

        assertThat(testReview.getResource()).isEqualTo("resource");
        assertThat(testReview.getRequestedBy()).isEqualTo(user.getId());
        assertThat(testReview.getComments().size()).isOne();

        ReviewRating reviewRating = testReview.getComments().stream().findFirst().get();

        assertThat(reviewRating.getStatus()).isEqualTo("Review improved");
        assertThat(reviewRating.getReview().getId()).isEqualTo(testReview.getId());
        assertThat(reviewRating.getUser().getEmail()).isEqualTo("admin@localhost");
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    public void rewardBadgeTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("admin@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));

        List<Review> reviewList = reviewRepository.findAll();

        Review testReview = reviewList.get(0);

        ReviewDTO reviewDTO = reviewService.reviewToDTO(testReview);

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/review/rewardBadge")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(reviewDTO))
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        reviewList = reviewRepository.findAll();

        testReview = reviewList.get(0);

        assertThat(testReview.getResource()).isEqualTo("resource");

        ReviewRating reviewRating = testReview.getComments().stream().findFirst().get();

        assertThat(reviewRating.getStatus()).isEqualTo("Review completed");
        assertThat(reviewRating.getReview().getId()).isEqualTo(testReview.getId());
        assertThat(reviewRating.getUser().getEmail()).isEqualTo("admin@localhost");

        Optional<Statistics> statistics = statisticsRepository.findByExerciseID(testReview.getResourceID());

        assertThat(statistics.isPresent()).isTrue();
        assertThat(statistics.get().getBadgeRewarded()).isTrue();
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER, username = "user")
    public void changeStatusTest() throws Exception {
        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders
                    .post("/api/review/changeStatus")
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
                    .content(TestUtil.convertObjectToJsonBytes(true))
            )
            .andReturn();

        User user = userService.getUserByLogin("user").get();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);
        assertThat(user.isReviewingEnabled()).isTrue();

        result =
            mockMvc
                .perform(
                    MockMvcRequestBuilders
                        .post("/api/review/changeStatus")
                        .with(csrf().asHeader())
                        .contentType(MediaType.APPLICATION_JSON)
                        .content(TestUtil.convertObjectToJsonBytes(false))
                )
                .andReturn();

        user = userService.getUserByLogin("user").get();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);
        assertThat(user.isReviewingEnabled()).isFalse();
    }

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.ADMIN, username = "admin")
    public void allWithStatisticsTest() throws Exception {
        List<String> reviewers = new ArrayList<>();
        reviewers.add("admin@localhost");

        reviewService.createReview(new ReviewRequest("resource", "codeability.at", "1", reviewers));
        reviewService.createReview(new ReviewRequest("resource1", "codeability1.at", "2", reviewers));
        reviewService.createReview(new ReviewRequest("resource2", "codeability2.at", "3", reviewers));

        List<Review> reviewList = reviewRepository.findAll();

        reviewList.forEach(review -> {
            reviewService.rewardBadge(review.getId());
        });

        MvcResult result = mockMvc
            .perform(
                MockMvcRequestBuilders.get("/api/review/allWithStatistics").with(csrf().asHeader()).contentType(MediaType.APPLICATION_JSON)
            )
            .andReturn();

        assertThat(result.getResponse().getStatus()).isEqualTo(200);

        String str = result.getResponse().getContentAsString();

        List<ReviewDTO> reviewDTOs = new ObjectMapper().readValue(str, new TypeReference<List<ReviewDTO>>() {});

        reviewDTOs.forEach(reviewDTO -> {
            assertThat(reviewDTO.getBadgeRewarded()).isTrue();
        });
    }
}
